@@ -20,6 +20,7 @@ Support for running 'shell commands'
 from __future__ import absolute_import
 from __future__ import print_function
 from future.builtins import range
+from future.utils import PY3
 from future.utils import iteritems
 from future.utils import string_types
 from future.utils import text_type
@@ -32,6 +33,7 @@ import stat
 import subprocess
 import sys
 import traceback
+from codecs import getincrementaldecoder
 from collections import deque
 from tempfile import NamedTemporaryFile
 
@@ -47,17 +49,20 @@ from twisted.python.win32 import quoteArguments
 
 from buildbot_worker import util
 from buildbot_worker.compat import bytes2NativeString
+from buildbot_worker.compat import bytes2unicode
+from buildbot_worker.compat import unicode2bytes
 from buildbot_worker.exceptions import AbandonChain
 
 if runtime.platformType == 'posix':
     from twisted.internet.process import Process
 
 
-def win32_batch_quote(cmd_list):
+def win32_batch_quote(cmd_list, unicode_encoding='utf-8'):
     # Quote cmd_list to a string that is suitable for inclusion in a
     # Windows batch file. This is not quite the same as quoting it for the
     # shell, as cmd.exe doesn't support the %% escape in interactive mode.
     def escape_arg(arg):
+        arg = bytes2NativeString(arg, unicode_encoding)
         arg = quoteArguments([arg])
         # escape shell special characters
         arg = re.sub(r'[@()^"<>&|]', r'^\g<0>', arg)
@@ -67,7 +72,7 @@ def win32_batch_quote(cmd_list):
     return ' '.join(map(escape_arg, cmd_list))
 
 
-def shell_quote(cmd_list, unicode_encoding='utf8'):
+def shell_quote(cmd_list, unicode_encoding='utf-8'):
     # attempt to quote cmd_list such that a shell will properly re-interpret
     # it.  The pipes module is only available on UNIX; also, the quote
     # function is undocumented (although it looks like it will be documented
@@ -77,19 +82,20 @@ def shell_quote(cmd_list, unicode_encoding='utf8'):
     # So:
     #  - use pipes.quote on UNIX, handling '' as a special case
     #  - use our own custom function on Windows
-    cmd_list = bytes2NativeString(cmd_list, unicode_encoding)
+    if isinstance(cmd_list, bytes):
+        cmd_list = bytes2unicode(cmd_list, unicode_encoding)
 
     if runtime.platformType == 'win32':
-        return win32_batch_quote(cmd_list)
-    else:
-        import pipes
+        return win32_batch_quote(cmd_list, unicode_encoding)
 
-        def quote(e):
-            if not e:
-                return '""'
-            e = bytes2NativeString(e, unicode_encoding)
-            return pipes.quote(e)
-        return " ".join([quote(e) for e in cmd_list])
+    import pipes  # only available on unix
+
+    def quote(e):
+        if not e:
+            return u'""'
+        e = bytes2unicode(e, unicode_encoding)
+        return pipes.quote(e)
+    return u" ".join([quote(e) for e in cmd_list])
 
 
 class LogFileWatcher(object):
@@ -100,7 +106,7 @@ class LogFileWatcher(object):
         self.name = name
         self.logfile = logfile
 
-        log.msg("LogFileWatcher created to watch %s" % logfile)
+        log.msg("LogFileWatcher created to watch {0}".format(logfile))
         # we are created before the ShellCommand starts. If the logfile we're
         # supposed to be watching already exists, record its size and
         # ctime/mtime so we can tell when it starts to change.
@@ -180,9 +186,13 @@ class RunProcessPP(protocol.ProcessProtocol):
 
     def __init__(self, command):
         self.command = command
-        self.pending_stdin = ""
+        self.pending_stdin = b""
         self.stdin_finished = False
         self.killed = False
+        decoderFactory = getincrementaldecoder(
+            self.command.builder.unicode_encoding)
+        self.stdoutDecode = decoderFactory(errors='replace')
+        self.stderrDecode = decoderFactory(errors='replace')
 
     def setStdin(self, data):
         assert not self.connected
@@ -194,8 +204,8 @@ class RunProcessPP(protocol.ProcessProtocol):
 
         if self.command.useProcGroup:
             if self.debug:
-                log.msg(" recording pid %d as subprocess pgid"
-                        % (self.transport.pid,))
+                log.msg(" recording pid {0} as subprocess pgid".format(
+                    self.transport.pid))
             self.transport.pgid = self.transport.pid
 
         if self.pending_stdin:
@@ -209,16 +219,14 @@ class RunProcessPP(protocol.ProcessProtocol):
     def outReceived(self, data):
         if self.debug:
             log.msg("RunProcessPP.outReceived")
-        data = bytes2NativeString(
-                   data, self.command.builder.unicode_encoding)
-        self.command.addStdout(data)
+        decodedData = self.stdoutDecode.decode(data)
+        self.command.addStdout(decodedData)
 
     def errReceived(self, data):
         if self.debug:
             log.msg("RunProcessPP.errReceived")
-        data = bytes2NativeString(
-                   data, self.command.builder.unicode_encoding)
-        self.command.addStderr(data)
+        decodedData = self.stderrDecode.decode(data)
+        self.command.addStderr(decodedData)
 
     def processEnded(self, status_object):
         if self.debug:
@@ -251,7 +259,6 @@ class RunProcess(object):
     shell.
     """
 
-    notreally = False
     BACKUP_TIMEOUT = 5
     interruptSignal = "KILL"
     CHUNK_LIMIT = 128 * 1024
@@ -278,7 +285,7 @@ class RunProcess(object):
                  sendStdout=True, sendStderr=True, sendRC=True,
                  timeout=None, maxTime=None, sigtermTime=None,
                  initialStdin=None, keepStdout=False, keepStderr=False,
-                 logEnviron=True, logfiles={}, usePTY=False,
+                 logEnviron=True, logfiles=None, usePTY=False,
                  useProcGroup=True):
         """
 
@@ -293,6 +300,8 @@ class RunProcess(object):
         @param useProcGroup: (default True) use a process group for non-PTY
             process invocations
         """
+        if logfiles is None:
+            logfiles = {}
 
         self.builder = builder
         if isinstance(command, list):
@@ -361,7 +370,7 @@ class RunProcess(object):
                 if v is not None:
                     if not isinstance(v, string_types):
                         raise RuntimeError("'env' values must be strings or "
-                                           "lists; key '%s' is incorrect" % (key,))
+                                           "lists; key '{0}' is incorrect".format(key))
                     newenv[key] = p.sub(subst, v)
 
             self.environ = newenv
@@ -422,7 +431,7 @@ class RunProcess(object):
             self.logFileWatchers.append(w)
 
     def __repr__(self):
-        return "<%s '%s'>" % (self.__class__.__name__, self.fake_command)
+        return "<{0} '{1}'>".format(self.__class__.__name__, self.fake_command)
 
     def sendStatus(self, status):
         self.builder.sendUpdate(status)
@@ -451,16 +460,11 @@ class RunProcess(object):
         if not os.path.isdir(self.workdir):
             os.makedirs(self.workdir)
         log.msg("RunProcess._startCommand")
-        if self.notreally:
-            self._addToBuffers('header', "command '%s' in dir %s" %
-                               (self.fake_command, self.workdir))
-            self._addToBuffers('header', "(not really)\n")
-            self.finished(None, 0)
-            return
 
         self.pp = RunProcessPP(self)
 
         self.using_comspec = False
+        self.command = unicode2bytes(self.command, encoding=self.builder.unicode_encoding)
         if isinstance(self.command, bytes):
             if runtime.platformType == 'win32':
                 # allow %COMSPEC% to have args
@@ -480,8 +484,10 @@ class RunProcess(object):
             # So, for .exe's that we have absolute paths to, we can call directly
             # Otherwise, we should run under COMSPEC (usually cmd.exe) to
             # handle path searching, etc.
-            if runtime.platformType == 'win32' and not \
-                    (self.command[0].lower().endswith(".exe") and os.path.isabs(self.command[0])):
+            if (runtime.platformType == 'win32' and
+                not (bytes2unicode(self.command[0],
+                     self.builder.unicode_encoding).lower().endswith(".exe") and
+                     os.path.isabs(self.command[0]))):
                 # allow %COMSPEC% to have args
                 argv = os.environ['COMSPEC'].split()
                 if '/c' not in argv:
@@ -494,7 +500,7 @@ class RunProcess(object):
             # isn't perfect
             display = shell_quote(self.fake_command, self.builder.unicode_encoding)
 
-        display = bytes2NativeString(display, self.builder.unicode_encoding)
+        display = bytes2unicode(display, self.builder.unicode_encoding)
 
         # $PWD usually indicates the current directory; spawnProcess may not
         # update this value, though, so we set it explicitly here.  This causes
@@ -504,52 +510,53 @@ class RunProcess(object):
 
         # self.stdin is handled in RunProcessPP.connectionMade
 
-        log.msg(" " + display)
-        self._addToBuffers('header', display + "\n")
+        log.msg(u" " + display)
+        self._addToBuffers(u'header', display + u"\n")
 
         # then comes the secondary information
-        msg = " in dir %s" % (self.workdir,)
+        msg = u" in dir {0}".format(self.workdir)
         if self.timeout:
             if self.timeout == 1:
-                unit = "sec"
+                unit = u"sec"
             else:
-                unit = "secs"
-            msg += " (timeout %d %s)" % (self.timeout, unit)
+                unit = u"secs"
+            msg += u" (timeout {0} {1})".format(self.timeout, unit)
         if self.maxTime:
             if self.maxTime == 1:
-                unit = "sec"
+                unit = u"sec"
             else:
-                unit = "secs"
-            msg += " (maxTime %d %s)" % (self.maxTime, unit)
-        log.msg(" " + msg)
-        self._addToBuffers('header', msg + "\n")
+                unit = u"secs"
+            msg += u" (maxTime {0} {1})".format(self.maxTime, unit)
+        log.msg(u" " + msg)
+        self._addToBuffers(u'header', msg + u"\n")
 
-        msg = " watching logfiles %s" % (self.logfiles,)
+        msg = " watching logfiles {0}".format(self.logfiles)
         log.msg(" " + msg)
-        self._addToBuffers('header', msg + "\n")
+        self._addToBuffers('header', msg + u"\n")
 
         # then the obfuscated command array for resolving unambiguity
-        msg = " argv: %s" % (self.fake_command,)
-        log.msg(" " + msg)
-        self._addToBuffers('header', msg + "\n")
+        msg = u" argv: {0}".format(self.fake_command)
+        log.msg(u" " + msg)
+        self._addToBuffers('header', msg + u"\n")
 
         # then the environment, since it sometimes causes problems
         if self.logEnviron:
-            msg = " environment:\n"
+            msg = u" environment:\n"
             env_names = sorted(self.environ.keys())
             for name in env_names:
-                msg += "  %s=%s\n" % (name, self.environ[name])
-            log.msg(" environment:\n%s" % (pprint.pformat(self.environ),))
-            self._addToBuffers('header', msg)
+                msg += u"  {0}={1}\n".format(bytes2unicode(name, encoding=self.builder.unicode_encoding),
+                                             bytes2unicode(self.environ[name], encoding=self.builder.unicode_encoding))
+            log.msg(u" environment:\n{0}".format(pprint.pformat(self.environ)))
+            self._addToBuffers(u'header', msg)
 
         if self.initialStdin:
-            msg = " writing %d bytes to stdin" % len(self.initialStdin)
-            log.msg(" " + msg)
-            self._addToBuffers('header', msg + "\n")
+            msg = u" writing {0} bytes to stdin".format(len(self.initialStdin))
+            log.msg(u" " + msg)
+            self._addToBuffers(u'header', msg + u"\n")
 
-        msg = " using PTY: %s" % bool(self.usePTY)
-        log.msg(" " + msg)
-        self._addToBuffers('header', msg + "\n")
+        msg = u" using PTY: {0}".format(bool(self.usePTY))
+        log.msg(u" " + msg)
+        self._addToBuffers(u'header', msg + u"\n")
 
         # put data into stdin and close it, if necessary.  This will be
         # buffered until connectionMade is called
@@ -579,10 +586,12 @@ class RunProcess(object):
         for w in self.logFileWatchers:
             w.start()
 
-    def _spawnProcess(self, processProtocol, executable, args=(), env={},
+    def _spawnProcess(self, processProtocol, executable, args=(), env=None,
                       path=None, uid=None, gid=None, usePTY=False, childFDs=None):
         """private implementation of reactor.spawnProcess, to allow use of
         L{ProcGroupProcess}"""
+        if env is None:
+            env = {}
 
         # use the ProcGroupProcess class, if available
         if runtime.platformType == 'posix':
@@ -594,22 +603,30 @@ class RunProcess(object):
         if self.using_comspec:
             return self._spawnAsBatch(processProtocol, executable, args, env,
                                       path, usePTY=usePTY)
-        else:
-            return reactor.spawnProcess(processProtocol, executable, args, env,
-                                        path, usePTY=usePTY)
+        return reactor.spawnProcess(processProtocol, executable, args, env,
+                                    path, usePTY=usePTY)
 
     def _spawnAsBatch(self, processProtocol, executable, args, env,
                       path, usePTY):
         """A cheat that routes around the impedance mismatch between
         twisted and cmd.exe with respect to escaping quotes"""
 
-        tf = NamedTemporaryFile(dir='.', suffix=".bat", delete=False)
-        # echo off hides this cheat from the log files.
-        tf.write("@echo off\n")
-        if isinstance(self.command, string_types):
-            tf.write(self.command)
+        # NamedTemporaryFile differs in PY2 and PY3.
+        # In PY2, it needs encoded str and its encoding cannot be specified.
+        # In PY3, it needs str which is unicode and its encoding can be specified.
+        if PY3:
+            tf = NamedTemporaryFile(mode='w+', dir='.', suffix=".bat",
+                                    delete=False, encoding=self.builder.unicode_encoding)
         else:
-            tf.write(win32_batch_quote(self.command))
+            tf = NamedTemporaryFile(mode='w+', dir='.', suffix=".bat",
+                                    delete=False)
+
+        # echo off hides this cheat from the log files.
+        tf.write(u"@echo off\n")
+        if isinstance(self.command, (string_types, bytes)):
+            tf.write(bytes2NativeString(self.command, self.builder.unicode_encoding))
+        else:
+            tf.write(win32_batch_quote(self.command, self.builder.unicode_encoding))
         tf.close()
 
         argv = os.environ['COMSPEC'].split()  # allow %COMSPEC% to have args
@@ -641,9 +658,9 @@ class RunProcess(object):
         """
         retval = {}
         for logname in msg:
-            data = ""
+            data = u""
             for m in msg[logname]:
-                m = bytes2NativeString(m, self.builder.unicode_encoding)
+                m = bytes2unicode(m, self.builder.unicode_encoding)
                 data += m
             if isinstance(logname, tuple) and logname[0] == 'log':
                 retval['log'] = (logname[1], data)
@@ -697,7 +714,7 @@ class RunProcess(object):
             # Chunkify the log data to make sure we're not sending more than
             # CHUNK_LIMIT at a time
             for chunk in self._chunkForSend(data):
-                if len(chunk) == 0:
+                if not chunk:
                     continue
                 logdata.append(chunk)
                 msg_size += len(chunk)
@@ -760,7 +777,7 @@ class RunProcess(object):
 
     def finished(self, sig, rc):
         self.elapsedTime = util.now(self._reactor) - self.startTime
-        log.msg("command finished with signal %s, exit code %s, elapsedTime: %0.6f" % (
+        log.msg("command finished with signal {0}, exit code {1}, elapsedTime: {2:0.6f}".format(
             sig, rc, self.elapsedTime))
         for w in self.logFileWatchers:
             # this will send the final updates
@@ -771,37 +788,38 @@ class RunProcess(object):
         if self.sendRC:
             if sig is not None:
                 self.sendStatus(
-                    {'header': "process killed by signal %d\n" % sig})
+                    {'header': "process killed by signal {0}\n".format(sig)})
             self.sendStatus({'rc': rc})
-        self.sendStatus({'header': "elapsedTime=%0.6f\n" % self.elapsedTime})
+        self.sendStatus({'header': "elapsedTime={0:0.6f}\n".format(self.elapsedTime)})
         self._cancelTimers()
         d = self.deferred
         self.deferred = None
         if d:
             d.callback(rc)
         else:
-            log.msg("Hey, command %s finished twice" % self)
+            log.msg("Hey, command {0} finished twice".format(self))
 
     def failed(self, why):
         self._sendBuffers()
-        log.msg("RunProcess.failed: command failed: %s" % (why,))
+        log.msg("RunProcess.failed: command failed: {0}".format(why))
         self._cancelTimers()
         d = self.deferred
         self.deferred = None
         if d:
             d.errback(why)
         else:
-            log.msg("Hey, command %s finished twice" % self)
+            log.msg("Hey, command {0} finished twice".format(self))
 
     def doTimeout(self):
         self.ioTimeoutTimer = None
-        msg = "command timed out: %d seconds without output running %s" % (
-            self.timeout, self.fake_command)
+        msg = (
+            "command timed out: {0} seconds without output running {1}".format(
+            self.timeout, self.fake_command))
         self.kill(msg)
 
     def doMaxTimeout(self):
         self.maxTimeoutTimer = None
-        msg = "command timed out: %d seconds elapsed running %s" % (
+        msg = "command timed out: {0} seconds elapsed running {1}".format(
             self.maxTime, self.fake_command)
         self.kill(msg)
 
@@ -846,62 +864,59 @@ class RunProcess(object):
             sig = getattr(signal, "SIG" + interruptSignal, None)
 
             if sig is None:
-                log.msg("signal module is missing SIG%s" % interruptSignal)
+                log.msg("signal module is missing SIG{0}".format(interruptSignal))
             elif not hasattr(os, "kill"):
                 log.msg("os module is missing the 'kill' function")
             elif self.process.pgid is None:
                 log.msg("self.process has no pgid")
             else:
-                log.msg("trying to kill process group %d" %
-                        (self.process.pgid,))
+                log.msg("trying to kill process group {0}".format(
+                        self.process.pgid))
                 try:
-                    os.kill(-self.process.pgid, sig)
-                    log.msg(" signal %s sent successfully" % sig)
+                    os.killpg(self.process.pgid, sig)
+                    log.msg(" signal {0} sent successfully".format(sig))
                     self.process.pgid = None
                     hit = 1
                 except OSError:
-                    log.msg('failed to kill process group (ignored): %s' %
-                            (sys.exc_info()[1],))
+                    log.msg('failed to kill process group (ignored): {0}'.format(
+                            (sys.exc_info()[1])))
                     # probably no-such-process, maybe because there is no process
                     # group
-                    pass
 
         elif runtime.platformType == "win32":
             if interruptSignal is None:
                 log.msg("interruptSignal==None, only pretending to kill child")
             elif self.process.pid is not None:
                 if interruptSignal == "TERM":
-                    log.msg("using TASKKILL PID /T to kill pid %s" %
-                            self.process.pid)
+                    log.msg("using TASKKILL PID /T to kill pid {0}".format(
+                            self.process.pid))
                     subprocess.check_call(
-                        "TASKKILL /PID %s /T" % self.process.pid)
-                    log.msg("taskkill'd pid %s" % self.process.pid)
+                        "TASKKILL /PID {0} /T".format(self.process.pid))
+                    log.msg("taskkill'd pid {0}".format(self.process.pid))
                     hit = 1
                 elif interruptSignal == "KILL":
-                    log.msg("using TASKKILL PID /F /T to kill pid %s" %
-                            self.process.pid)
+                    log.msg("using TASKKILL PID /F /T to kill pid {0}".format(
+                            self.process.pid))
                     subprocess.check_call(
-                        "TASKKILL /F /PID %s /T" % self.process.pid)
-                    log.msg("taskkill'd pid %s" % self.process.pid)
+                        "TASKKILL /F /PID {0} /T".format(self.process.pid))
+                    log.msg("taskkill'd pid {0}".format(self.process.pid))
                     hit = 1
 
         # try signalling the process itself (works on Windows too, sorta)
         if not hit:
             try:
-                log.msg("trying process.signalProcess('%s')" %
-                        (interruptSignal,))
+                log.msg("trying process.signalProcess('{0}')".format(
+                        interruptSignal))
                 self.process.signalProcess(interruptSignal)
-                log.msg(" signal %s sent successfully" % (interruptSignal,))
+                log.msg(" signal {0} sent successfully".format(interruptSignal))
                 hit = 1
             except OSError:
                 log.err("from process.signalProcess:")
                 # could be no-such-process, because they finished very recently
-                pass
             except error.ProcessExitedAlready:
                 log.msg("Process exited already - can't kill")
                 # the process has already exited, and likely finished() has
                 # been called already or will be called shortly
-                pass
 
         return hit
 
@@ -939,7 +954,8 @@ class RunProcess(object):
         self.failed(RuntimeError(signalName + " failed to kill process"))
 
     def _cancelTimers(self):
-        for timerName in ('ioTimeoutTimer', 'killTimer', 'maxTimeoutTimer', 'sendBuffersTimer', 'sigtermTimer'):
+        for timerName in ('ioTimeoutTimer', 'killTimer', 'maxTimeoutTimer',
+                          'sendBuffersTimer', 'sigtermTimer'):
             timer = getattr(self, timerName, None)
             if timer:
                 timer.cancel()
